% Convergence model for diffusion LMS algorithm in multitask environments
% (Figure 2-5)
%
% Reference:
% Jie Chen, Cedric Richard, Ali H. Sayed. 
% Diffusion LMS over multitask networks
% IEEE Transactions on Signal Processing, 2015
% Also avalable at http://arxiv.org/pdf/1404.6813v2.pdf.
%
% Version Mar. 2015
%
% By: Jie Chen
% dr.jie.chen@ieee.org
% http://www.jie-chen.com


clc, clear

%% Simulation parameter
RLZ = 20;             % Realization number (we put 20 here for illustration)
T = 1000;             % Run time
N = 8;                % Node number (associated wit Gene_topology2)
L = 2;                % Regression dimension

mu0 = 0.01;
mu = mu0*ones(N,1);    % Step-sizes (here all identical)


%% Topology generation

% Connection matrix: CM; Cluster matrix: ClsMask
Gene_topology2


%% Stochastic matrices

% Glboal cost decomposition matrix

% - Uniform matrix without cluster consideration:
C = diag(1./sum(CM))*CM;       
% - Without gradient exchage (non-coorpative LMS)
%C = eye(N);
        

% Combination matrix  %(note here A is the A' in the paper)
 A = diag(1./sum(CM))*CM;      
 %A = eye(N);


% Signal power
var_x = (rand(N,1)*0.4-0.2) + 1;
var_z = 0.1*((rand(N,1)*0.4-0.2) + 2);

% system coefficients  
r0 = 0.01;          % radius
c = [1,-0.5]';      % center
phi0 = pi/8; 
stds =0.001;       % standard deviation 
fig = 1;
[ws0, ws]=Generate_circweights(N,r0,c,phi0,stds,fig);



% Node coefficient
W = zeros(L*N, T);
U = zeros(L*N, 1);
Wav = W;
MSD = zeros(T,1);
MSDav = MSD;

% Simulation
X = zeros(L,N);
dN = zeros(N,1);

display('---Starting MC simulation---')

for rlz = 1 : RLZ
    
    
    rlz

for t = 2 : T
    
     ws_xi = stds*randn(N*L,1);
     for n = 1 : N
         X(:,n) = sqrt(var_x(n))*randn(L,1);
         wsn = ws((n-1)*L+1: n*L);         
         dN(n) = X(:,n)'*(wsn+ws_xi((n-1)*L+1:n*L)) + sqrt(var_z(n))*randn;
     end
     
     for n = 1 : N
         Gd = zeros(L,1);
         sw = Gd;
         for m = 1 : N
            if C(m,n)>0
                 Gd = Gd + C(m,n) * (dN(m) - X(:,m)'*W((n-1)*L+1:n*L,t-1))*X(:,m);
            end
         end
            U((n-1)*L+1:n*L) = W((n-1)*L+1:n*L,t-1) + mu(n)*Gd;
     end
     
     for n = 1 : N
         W((n-1)*L+1:n*L,t) = sum(reshape(U,L,[])*diag(A(n,:)),2);    
     end
    MSD(t) = (W(:,t)-ws-ws_xi)'*(W(:,t)-ws-ws_xi)/N ;
end
        Wav = Wav+W;
        MSDav = MSDav + MSD;
end

Wav = Wav/RLZ;%-ws*ones(1,T);
MSDav = MSDav/RLZ;

figure
plot(Wav','b'), hold on
%plot((ws*ones(1,T))','r')


display('--- Analytical model computation ----')
% Analytical model
A=A';    %(note in the MC part A is the A' in the paper)


AI  = kron(A,eye(L));
Q = zeros(N*L);%eye(N*L) - kron(P,eye(L));
CI = kron(C,eye(L));
Rb = kron(var_x,eye(L))';
HR = bdiag(Rb*CI);
MU = bdiag(kron(mu, eye(L))');


Evt = zeros(L*N,T);
Evt(:,1) = -ws;

%Ev_inf = - mu *lambda*inv(AI'*(eye(N*L)-mu*(HR+lambda*Q)) - eye(N*L))*AI'*Q*ws;

B= AI'*(eye(N*L)-MU*HR);
K = kron(B' , B');
G = AI'*MU*CI'*bdiag(kron(var_x.*var_z,eye(L))')*CI*MU*AI;
g = reshape(G',[],1);
sig_msd = inv(eye((N*L)^2)-K)*reshape(eye(N*L),[],1)/N;
Sig_msd = reshape(sig_msd,N*L,N*L);


% \bP = \bw^* \,(\cb{1}_N\otimes \bI_L)^\top - (\cb{1}_N\otimes \bI_L) \, \bw^*
wss = reshape(ws,L,N);

P = kron(ones(N,1),eye(L))*wss - ws*ones(1,N) ;
P = kron(C,ones(L,1)).*P;
hu = reshape(Rb*P,[],1);
 
r = AI'*MU*hu - (AI'-eye(N*L))*ws;

Ev_inf = -inv(eye(N*L)-B)*r;

f_inf = r'*Sig_msd*r - 2*(r'*Sig_msd*B*Ev_inf);
f_inf = r'*Sig_msd*r - (r'*Sig_msd*B*Ev_inf+ (B*Ev_inf)'*Sig_msd*r);

MSD_inf = trace(G*Sig_msd) + f_inf+L*stds^2;
10*log10(MSD_inf)

MSDt = zeros(T,1);
MSDt(1) = ws'*ws/N;


sigI = reshape(eye((N*L)),[],1);


v0 = -ws;
MecK = (eye((N*L)^2)-K);
v0v0 = kron(v0',v0');
rr = kron(r',r');
Kt = eye(size(K));
sKBv = zeros(size(sigI))';%kron((B*v0)',r')*(K-eye(size(K)));
for t = 2 : T
     Evt(:,t) = B*Evt(:,t-1) - r;
     
    
   %  MSDt(t) = MSDt(t-1) + 1/N*(mu^2*g'*(Kt*sigI) + mu^2*lambda^2*rr*(Kt*sigI) -2* mu*lambda*kron((B*Evt(:,t-1))',r')*(Kt*sigI)-v0v0*MecK*sigI);
      Sig = reshape(Kt*sigI, N*L,N*L);
      MSDt(t) = MSDt(t-1) + 1/N*(trace(G*Sig) + r'*Sig*r -2*(sKBv*sigI + r'*B*Evt(:,t-1))-v0v0*MecK*sigI);
     Kt = K*Kt;
     MecK = MecK*K;
     sKBv = sKBv*K + kron((B*Evt(:,t-1))',r')*(K-eye(size(K)));
end

plot((Evt+ws*ones(1,T))','m','linewidth',1)
xlabel('Iteration $n$','fontsize',14,'interpreter', 'latex')
ylabel('Weights $w_i(n)$','fontsize',14,'interpreter', 'latex')
%norm(- mu *lambda*inv(AI'*(eye(N*L)-mu*(HR+lambda*Q)) - eye(N*L))*AI'*Q*ws)

%norm(Wav(:,T)-ws)

figure, plot(10*log10(MSDav),'linewidth',2), hold on
plot(10*log10(MSDt+L*stds^2),'m','linewidth',2)
%plot(10*log10(MSDtNon),'k')

plot(ones(size(MSDav))*10*log10(MSD_inf),'r','linewidth',2)
xlabel('Iteration $n$','fontsize',14,'interpreter', 'latex')
ylabel('MSD','fontsize',14,'interpreter', 'latex')
grid on
