% Adaptive clustering in stationary environment 
% (Section V B-1, Fig 7)
%
% Reference:
% Jie Chen, Cedric Richard, Ali H. Sayed. 
% Diffusion LMS over multitask networks
% IEEE Transactions on Signal Processing, 2015
% Also avalable at http://arxiv.org/pdf/1404.6813v2.pdf.
%
% Verision Mar. 2015
%
% By: Jie Chen
% dr.jie.chen@ieee.org
% http://www.jie-chen.com

clc, clear

%% Simulation control parameters
% Realization numbers
RLZ = 10;
% run time 
T = 1000;
% step size
mu = 0.01;
% combine strategy
% 0:  No-combination (Non-cooperative LMS)
% 1:  Neighborhood uniform combination  (convetional diffusion LMS)
% 2:  Perfect clustering combination  (Known clusters)
% 3:  Sayed combination (ref [41])
% 4:  Proposed
% 5:  Proposed with C = A';
A_STR =4;

% Gradient strategy (GS = 0 for normalized)
GS = 0;


%% Topology generation

% N = 16, Q = 4; 
% Connection matrix: CM; Cluster matrix: ClsMask
Gene_topology1


%% Stochastic matrices

% Glboal cost decomposition matrix
C = eye(N);
        

% Combination matrix

switch A_STR
    case 0
        % - Without gradient exchage:
        A = eye(N);
    case 1
        % - Uniform matrix without cluster consideration:
        A = diag(1./sum(CM))*CM;      
    case 2
        % - Uniform matrix with perfect clustering: 
        A = diag(1./sum(CM.*ClsMask))*CM.*ClsMask;   
    case {3,4,5}
        % Intialized with the identitiy matrix
        A = eye(N);
    otherwise
        error('No combination strategy is assigned')
end

 



%% Signal and system

% Signal power
var_x = (rand(N,1)*0.4-0.2) + 1;
var_z = 0.1*((rand(N,1)*0.4-0.2) + 2);


% system coefficients
L = 2;
% Coefficients w.r.t. clusters L * Q
ws =  [0.5,  0.6,  0.3, -0.8
      -0.4, -0.2, -0.3, 0.5];
  
% Coefficients allocated to nodes L * N
wn = zeros(L, N);
wn(:, 1:4) = ws(:,1)*ones(1,4);
wn(:, 5:9) = ws(:,2)*ones(1,5);  
wn(:, 10:14) = ws(:,3)*ones(1,5);
wn(:, 15:16) = ws(:,4)*ones(1,2);
% vectorized coefficients
wnv =reshape(wn,[],1);


%% Distributed processing
% Initialization
% Node coefficient
W = zeros(L*N, T);
%load W
%W(:,1)=rand(L*N,T);

U = zeros(L*N, 1);
U2 = U;
UU = zeros(L*N, T);
Wav = W;
MSD = zeros(T,1);
MSDav = MSD;

% Simulation ATC algorithm
X = zeros(L,N);
dN = zeros(N,1);
Aav = 0;

for rlz = 1 : RLZ
rlz

if A_STR == 3 | A_STR == 4 |  A_STR == 5
    A = eye(N);
end

% Time evolution
Gam = zeros(N);
for t = 2 : T

    % Generate data on each node
    for n = 1 : N
        X(:,n) = sqrt(var_x(n))*randn(L,1);
        dN(n) = X(:,n)'*(wn(:,n)+0.0*randn(L,1)) + sqrt(var_z(n))*randn;
    end

    % Gradient exchage with C
    if A_STR == 5
        C = A;               %(note here A is the A' in the paper)
    else
        C = eye(N);
    end
    
  
    for n = 1 : N
        Gd = zeros(L,1);
        sw = Gd;
        for m = 1 : N
            if C(m,n)>0
                Gd = Gd + C(m,n) * (dN(m) - X(:,m)'*W((n-1)*L+1:n*L,t-1))*X(:,m);
            end
        end
        U((n-1)*L+1:n*L) = W((n-1)*L+1:n*L,t-1) + mu*Gd;
        Gd = X(:,n)*(dN(n) - X(:,n)'*U((n-1)*L+1:n*L));
        if GS == 1
           Gd = Gd /(norm(Gd)+0.1);
        end
        U2((n-1)*L+1:n*L) = U((n-1)*L+1:n*L)  + mu*Gd;
    end
    UU(:,t) = U;
      

    
    % Autonomous Combination strategy
    if A_STR == 3
        nu = 0.1;
        Wpre = reshape(W(:,t-1),L,N);
        Un = reshape(U,L,N);
        Normnew = (ones(N,1)*ones(1,L)*(Wpre.^2))' +ones(N,1)*ones(1,L)*(Un.^2) - 2* Wpre' * Un;
        Gam = (1-nu)*Gam + nu*Normnew;
        Gam2inv = 1./(Gam);
        A = diag(1./sum(Gam2inv.*CM,2)) *(Gam2inv.*CM);
    %    A = A';
    elseif A_STR == 4  |  A_STR == 5
        Un = reshape(U,L,N);
        for i = 1 : N
            Wopte =  U2((i-1)*L+1:i*L);
            Wdiff = Un(:,CM(i,:)>0)-Wopte*ones(1,sum(CM(i,:)>0));
            Wn = sum(Wdiff.^2);
            Wnd = 1./Wn;
            A(i,CM(i,:)>0) = Wnd/sum(Wnd);
        end
    end
    for n = 1 : N
        W((n-1)*L+1:n*L,t) = sum(reshape(U,L,[])*diag(A(n,:)),2);
    end
    MSD(t) = (W(:,t)-wnv)'*(W(:,t)-wnv)/N ;

end
Wav = Wav+W;
MSDav = MSDav + MSD;
end

Wav = Wav/RLZ-0*wnv*ones(1,T);
MSDav = MSDav/RLZ;

figure
plot(Wav(1:L*4,:)','b'), hold on
plot(Wav(L*4+1:L*9,:)','g'),
plot(Wav(L*9+1:L*14,:)','r'),
plot(Wav(L*14+1:end,:)','m'),

plot((reshape(ws,[],1)*ones(1,T))','k')

xlabel('Iteration $n$','fontsize',14,'interpreter', 'latex')
ylabel('Weights $w_i(n)$','fontsize',14,'interpreter', 'latex')


figure, plot(10*log10(MSDav),'linewidth',2), hold on
xlabel('Iteration $n$','fontsize',14,'interpreter', 'latex')
ylabel('MSD','fontsize',14,'interpreter', 'latex')
