% Adaptive clustering in non-stationary environment 
% (Section V B-2, Fig 10,11)
%
% Reference:
% Jie Chen, Cedric Richard, Ali H. Sayed. 
% Diffusion LMS over multitask networks
% IEEE Transactions on Signal Processing, 2015
% Also avalable at http://arxiv.org/pdf/1404.6813v2.pdf.
%
% Verision Mar. 2015
%
% By: Jie Chen
% dr.jie.chen@ieee.org
% http://www.jie-chen.com

clc, clear

%% Simulation control parameters
% Realization numbers
RLZ = 5;
% run time 
T1 = 1500;
T = 5500
% step size
mu = 0.01;
% combine strategy
% 0:  No-combination (Non-cooperative LMS)
% 1:  Neighborhood uniform combination  (convetional diffusion LMS)
% 2:  Perfect clustering combination  (Known clusters)
% 3:  Sayed combination (ref [41])
% 4:  Proposed
% 5:  Proposed with C = A';
A_STR =5;

% Gradient strategy (GS = 1 for normalized)
GS = 0;



%% Topology generation

% N = 16, Q = 4; 
% Connection matrix: CM; Cluster matrix: ClsMask
Gene_topology1


%% Stochastic matrices

% Glboal cost decomposition matrix

C = eye(N);
        

% Combination matrix

switch A_STR
    case 0
        % - Without gradient exchage:
        A = eye(N);
    case 1
        % - Uniform matrix without cluster consideration:
        A = diag(1./sum(CM))*CM;      
    case 2
        % - Uniform matrix with perfect clustering: 
        A = diag(1./sum(CM.*ClsMask))*CM.*ClsMask;   
    case {3,4,5}
        % - Centre bias combination (Sayed) 
        % Only initialization is given here
        A = zeros(N);
    otherwise
        error('No combination strategy is assigned')
end

 



%% Signal and system

% Signal power
var_x = (rand(N,1)*0.4-0.2) + 1;
var_z = 0.1*((rand(N,1)*0.4-0.2) + 2);


% system coefficients
L = 2;


% Phase 1:
wn1 = zeros(L, N);
ws1 = [0.3,  -0.2]';
wn1 = ws1*ones(1,N);

% phase 2;
wn2 = zeros(L, N);
ws2 =[0.5, 0.2; 
     -0.3, -0.1];
wn2(:, 1:9) = ws2(:,1)*ones(1,9);
wn2(:,10:16) = ws2(:,2)*ones(1,7);


% phase 3:
ws3 =  [0.4,  0.6,  0.3, 0.1
      -0.4, -0.2, -0.2, 0];
wn3 = zeros(L, N);
wn3(:, 1:4) = ws3(:,1)*ones(1,4);
wn3(:, 5:9) = ws3(:,2)*ones(1,5);  
wn3(:, 10:14) = ws3(:,3)*ones(1,5);
wn3(:, 15:16) = ws3(:,4)*ones(1,2);

% phase 4:
wn4 = zeros(L, N);
ws4 = [0.1,  -0.1]';
wn4 = ws4*ones(1,N);

% vectorized coefficients



%% Distributed processing
% Initialization
% Node coefficient
W = zeros(L*N, T);
%load W
%W(:,1)=W0;
U = zeros(L*N, 1);
U2 = U;
UU = zeros(L*N, T);
Wav = W;
MSD = zeros(T,1);
MSDav = MSD;

% Simulation ATC algorithm
X = zeros(L,N);
dN = zeros(N,1);




for rlz = 1 : RLZ
rlz

if A_STR == 3 | A_STR == 4 |  A_STR == 5
    A = eye(N);
end

% Time evolution
Gam = zeros(N);
for t = 2 : T
    
    
    if t < T1
        if t == 2
        wn = wn1;
        wnv =reshape(wn,[],1);
        end
        if t >= 1000
            rho = (t-1000)/500; 
            wn = wn1*(1-rho) + wn2*rho;
            wnv =reshape(wn,[],1);
        end
    elseif t < 2*T1
        if t == T1
        wn = wn2;
        wnv =reshape(wn,[],1);
        end
        if t >= 2500
            rho = (t-2500)/500; 
            wn = wn2*(1-rho) + wn3*rho;
            wnv =reshape(wn,[],1);
        end
    elseif t < 3*T1
        if t == 2*T1
        wn=wn3;
        wnv =reshape(wn,[],1);
        end
        if t >= 4000
            rho = (t-4000)/500; 
            wn = wn3*(1-rho) + wn4*rho;
            wnv =reshape(wn,[],1);
        end
    elseif t == 3*T1
        wn = wn4;
        wnv =reshape(wn,[],1);
    end
    
    

    % Generate data on each node
    for n = 1 : N
        X(:,n) = sqrt(var_x(n))*randn(L,1);
        dN(n) = X(:,n)'*(wn(:,n)+0.0*randn(L,1)) + sqrt(var_z(n))*randn;
    end

    % Gradient exchage with C
    if A_STR == 5
        C = A;               %(note here A is the A' in the paper)
    else
        C = eye(N);
    end
    
  
    for n = 1 : N
        Gd = zeros(L,1);
        sw = Gd;
        for m = 1 : N
            if C(m,n)>0
                Gd = Gd + C(m,n) * (dN(m) - X(:,m)'*W((n-1)*L+1:n*L,t-1))*X(:,m);
                if m == n
                    Gd0 =Gd;
                end
            end
        end
        U((n-1)*L+1:n*L) = W((n-1)*L+1:n*L,t-1) + mu*Gd;
        Gd = X(:,n)*(dN(n) - X(:,n)'*U((n-1)*L+1:n*L));
        if GS == 1
           Gd = Gd /(norm(Gd)+0.1);
        end
        U2((n-1)*L+1:n*L) = U((n-1)*L+1:n*L)  + mu*Gd;
    end
    UU(:,t) = U;
      

    
    % Autonomous Combination strategy
    if A_STR == 3
        nu = 0.1;
        Wpre = reshape(W(:,t-1),L,N);
        Un = reshape(U,L,N);
        Normnew = (ones(N,1)*ones(1,L)*(Wpre.^2))' +ones(N,1)*ones(1,L)*(Un.^2) - 2* Wpre' * Un;
        Gam = (1-nu)*Gam + nu*Normnew;
        Gam2inv = 1./(Gam);
        A = diag(1./sum(Gam2inv.*CM,2)) *(Gam2inv.*CM);
%        A = A';
    elseif A_STR == 4  |  A_STR == 5
        Un = reshape(U,L,N);
        for i = 1 : N
            Wopte =  U2((i-1)*L+1:i*L);%wn(:,i);
            Wdiff = Un(:,CM(i,:)>0)-Wopte*ones(1,sum(CM(i,:)>0));
            Wn = sum(Wdiff.^2);
            Wnd = 1./Wn;
            A(i,CM(i,:)>0) = Wnd/sum(Wnd);
        end
    end
    for n = 1 : N
        W((n-1)*L+1:n*L,t) = sum(reshape(U,L,[])*diag(A(n,:)),2);
    end
    MSD(t) = (W(:,t)-wnv)'*(W(:,t)-wnv)/N ;
end
Wav = Wav+W;
MSDav = MSDav + MSD;
end

Wav = Wav/RLZ-0*wnv*ones(1,T);
MSDav = MSDav/RLZ;

figure
plot(Wav(1:L*4,:)','b'), hold on
plot(Wav(L*4+1:L*9,:)','r'),
plot(Wav(L*9+1:L*14,:)','m'),
plot(Wav(L*14+1:end,:)','g'),

xlabel('Iteration $n$','fontsize',18,'interpreter', 'latex')
ylabel('Mean weight evolution','fontsize',18,'interpreter', 'latex')
ha=gca; set(ha,'fontsize',12)
xlim([1,5500]), ylim([-0.45,0.62]);



figure, plot(10*log10(MSDav)), hold on, grid on
xlabel('Iteration $n$','fontsize',18,'interpreter', 'latex')
ylabel('MSD (dB)','fontsize',18,'interpreter', 'latex')
ha=gca; set(ha,'fontsize',12)
xlim([1,5500])
